import multiprocessing
from pathlib import Path
from typing import List

from Cython.Build import cythonize
from Cython.Distutils.build_ext import new_build_ext as cython_build_ext
from setuptools import Extension, Distribution

SOURCE_DIR = Path("pybwa")
BUILD_DIR = Path("cython_build")
compile_args = []
link_args = []
include_dirs = ["bwa"]
libraries = ['m', 'z', 'pthread']
library_dirs=['bwa']
extra_objects = []
h_files = []
c_files = []
for root_dir in ["bwa", "pybwa"]:
    h_files.extend(str(x) for x in Path(root_dir).rglob("*.h"))
    c_files.extend(str(x) for x in Path(root_dir).rglob("*.c") if x.name not in ['example.c', 'main.c'])

libbwaindex_module = Extension(
    name='pybwa.libbwaindex',
    sources=['pybwa/libbwaindex.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
)

libbwaaln_module = Extension(
    name='pybwa.libbwaaln',
    sources=['pybwa/libbwaaln.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
)

libbwamem_module = Extension(
    name='pybwa.libbwamem',
    sources=['pybwa/libbwamem.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
)


def cythonize_helper(extension_modules: List[Extension]) -> List[Extension]:
    """Cythonize all Python extensions"""

    return cythonize(
        module_list=extension_modules,

        # Don't build in source tree (this leaves behind .c files)
        build_dir=BUILD_DIR,

        # Don't generate an .html output file. Would contain source.
        annotate=False,

        # Parallelize our build
        nthreads=multiprocessing.cpu_count() * 2,

        # Tell Cython we're using Python 3. Becomes default in Cython 3
        compiler_directives={"language_level": "3", 'embedsignature': True},

        # (Optional) Always rebuild, even if files untouched
        force=True,
    )

CLASSIFIERS = '''
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
'''


def build():
    # Collect and cythonize all files
    extension_modules = cythonize_helper([
        libbwaindex_module,
        libbwaaln_module,
        libbwamem_module
    ])

    # Use Setuptools to collect files
    distribution = Distribution({
        "name": "pybwa",
        'version': '0.0.1',
        'description': 'Python bindings for BWA',
        'long_description': __doc__,
        'long_description_content_type': 'text/x-rst',
        'author': 'Nils Homer',
        'author_email': 'nils@fulcrumgenomics.com',
        'license': 'MIT',
        'platforms': ['POSIX', 'UNIX', 'MacOS'],
        'classifiers': [_f for _f in CLASSIFIERS.split('\n') if _f],
        'url': 'https://github.com/fulcrumgenomics/pybwa',
        'packages': ['pybwa'],
        'package_dirs': {'pybwa': 'pybwa'},
        "ext_modules": extension_modules,
        "cmdclass": {
            "build_ext": cython_build_ext,
        },
    })

    # Grab the build_ext command and copy all files back to source dir.
    # Done so Poetry grabs the files during the next step in its build.
    build_ext_cmd = distribution.get_command_obj("build_ext")
    build_ext_cmd.ensure_finalized()
    # Set the value to 1 for "inplace", with the goal to build extensions
    # in build directory, and then copy all files back to the source dir
    # (under the hood, "copy_extensions_to_source" will be called after
    # building the extensions). This is done so Poetry grabs the files
    # during the next step in its build.
    build_ext_cmd.inplace = 1
    build_ext_cmd.run()


if __name__ == "__main__":
    build()



# cython: language_level=3

from pathlib import Path
from typing import List
from fgpyo.sequence import reverse_complement

from libc.stdint cimport uint8_t
from libc.stdlib cimport calloc, free
from libc.string cimport strncpy
from pysam import FastxRecord, AlignedSegment, qualitystring_to_array
from pybwa.libbwaindex cimport force_bytes
from pybwa.libbwaindex cimport BwaIndex


__all__ = [
    "BwaAlnOptions",
    "BwaAln",
]

cdef class BwaAlnOptions:
    """The container for options for :class:`pybwa.BwaAln`.
    
    Args:
        max_mismatches (int | None): :code:`-n <int>`
        max_gap_opens (int | None): :code:`-o <int>`
        max_gap_extensions (int | None): :code:`-e <int>`
        min_indel_to_end_distance (int | None): :code:`-i <int>`
        max_occurrences_for_extending_long_deletion (int | None): :code:`-d <int>`
        seed_length (int | None): :code:`-l <int>`
        max_mismatches_in_seed (int | None): :code:`-k <int>`
        mismatch_penalty (int | None): :code:`-M <int>`
        gap_open_penalty (int | None): :code:`-O <int>`
        gap_extension_penalty (int | None): :code:`-E <int>`
        stop_at_max_best_hits (int | None): :code:`-R <int>`
        max_hits (int | None): :code:`bwa samse -n <int>`
        log_scaled_gap_penalty (in | None): :code:`-L`
    """
    cdef gap_opt_t * _delegate

    _max_hits: int

    def __init__(self,
                 max_mismatches: int | None = None,
                 max_gap_opens: int | None = None,
                 max_gap_extensions: int | None = None,
                 min_indel_to_end_distance: int | None = None,
                 max_occurrences_for_extending_long_deletion: int | None = None,
                 seed_length: int | None = None,
                 max_mismatches_in_seed: int | None = None,
                 mismatch_penalty: int | None = None,
                 gap_open_penalty: int | None = None,
                 gap_extension_penalty: int | None = None,
                 stop_at_max_best_hits: int | None = None,
                 max_hits: int | None = 3,
                 log_scaled_gap_penalty: bool | None = None
                 ):
        if max_mismatches is not None:
            self.max_mismatches = max_mismatches
        if max_gap_extensions is not None:
            self.max_gap_extensions = max_gap_extensions
        if min_indel_to_end_distance is not None:
            self.min_indel_to_end_distance = min_indel_to_end_distance
        if max_occurrences_for_extending_long_deletion is not None:
            self.max_occurrences_for_extending_long_deletion = max_occurrences_for_extending_long_deletion
        if seed_length is not None:
            self.seed_length = seed_length
        if max_mismatches_in_seed is not None:
            self.max_mismatches_in_seed = max_mismatches_in_seed
        if mismatch_penalty is not None:
            self.mismatch_penalty = mismatch_penalty
        if gap_open_penalty is not None:
            self.gap_open_penalty = gap_open_penalty
        if gap_extension_penalty is not None:
            self.gap_extension_penalty = gap_extension_penalty
        if stop_at_max_best_hits is not None:
            self.stop_at_max_best_hits = stop_at_max_best_hits
        if max_hits is not None:
            self.max_hits = max_hits
        if log_scaled_gap_penalty is not None:
            self.log_scaled_gap_penalty = 1 if log_scaled_gap_penalty else 0

    def __cinit__(self):
        self._delegate = gap_init_opt()

    def __dealloc__(self):
        free(self._delegate)

    cdef gap_opt_t* gap_opt(self):
        """Returns the options struct to use with the bwa C library methods"""
        return self._delegate

    property max_mismatches:
        """:code:`bwa aln -n <int>`"""
        def __get__(self) -> int:
            return self._delegate.s_mm
        def __set__(self, value: int):
            self._delegate.s_mm = value

    property max_gap_opens:
        """:code:`bwa aln -o <int>`"""
        def __get__(self) -> int:
            return self._delegate.max_gapo
        def __set__(self, value: int):
           self._delegate.max_gapo = value

    property max_gap_extensions:
        """:code:`bwa aln -e <int>`"""
        def __get__(self) -> int:
            return self._delegate.max_gapo
        def __set__(self, value: int):
            self._delegate.max_gape = value
            if self._delegate.max_gape > 0:
                self._delegate.mode &= ~BWA_MODE_GAPE

    property min_indel_to_end_distance:
        """:code:`bwa aln -i <int>`"""
        def __get__(self) -> int:
            return self._delegate.indel_end_skip
        def __set__(self, value: int):
           self._delegate.indel_end_skip = value

    property max_occurrences_for_extending_long_deletion:
        """:code:`bwa aln -d <int>`"""
        def __get__(self) -> int:
            return self._delegate.max_del_occ
        def __set__(self, value: int):
           self._delegate.max_del_occ = value

    property seed_length:
        """:code:`bwa aln -l <int>`"""
        def __get__(self) -> int:
            return self._delegate.seed_len
        def __set__(self, value: int):
           self._delegate.seed_len = value

    property max_mismatches_in_seed:
        """:code:`bwa aln -k <int>`"""
        def __get__(self) -> int:
            return self._delegate.max_seed_diff
        def __set__(self, value: int):
           self._delegate.max_seed_diff = value

    property mismatch_penalty:
        """:code:`bwa aln -M <int>`"""
        def __get__(self) -> int:
            return self._delegate.s_mm
        def __set__(self, value: int):
           self._delegate.s_mm = value

    property gap_open_penalty:
        """:code:`bwa aln -O <int>`"""
        def __get__(self) -> int:
            return self._delegate.s_gapo
        def __set__(self, value: int):
           self._delegate.s_gapo = value

    property gap_extension_penalty:
        """:code:`bwa aln -E <int>`"""
        def __get__(self) -> int:
            return self._delegate.s_gape
        def __set__(self, value: int):
           self._delegate.s_gape = value

    property stop_at_max_best_hits:
        """:code:`bwa aln -R <int>`"""
        def __get__(self) -> int:
            return self._delegate.max_top2
        def __set__(self, value: int):
           self._delegate.max_top2 = value

    property max_hits:
        """:code:`bwa samse -n <int>`"""
        def __get__(self) -> int:
            return self._max_hits
        def __set__(self, value: int):
           self._max_hits = value

    property log_scaled_gap_penalty:
        """:code:`bwa aln -L`"""
        def __get__(self) -> int:
            return self._delegate.mode & BWA_MODE_LOGGAP > 0
        def __set__(self, value: int):
            if value:
                self._delegate.mode |= BWA_MODE_LOGGAP
            else:
                self._delegate.mode &= ~BWA_MODE_LOGGAP


cdef class BwaAln:
    """The class to align reads with :code:`bwa aln`."""

    cdef BwaIndex _index

    def __init__(self, prefix: str | Path | None = None, index: BwaIndex | None = None):
        """Constructs the :code:`bwa aln` aligner.

        One of `prefix` or `index` must be specified.

        Args:
            prefix: the path prefix for the BWA index (typically a FASTA)
            index: the index to use
        """
        if prefix is not None:
            assert Path(prefix).exists()
            self._index = BwaIndex(prefix=prefix)
        elif index is not None:
            self._index = index
        else:
            raise Exception("Either prefix or index must be given")

        bwase_initialize()

    def align(self, queries: List[FastxRecord] | List[str] | None = None, opt: BwaAlnOptions | None = None) -> List[AlignedSegment]:
        """Align one or more queries with `bwa aln`.

        Args:
            queries: the queries to align
            opt: the alignment options | None = None, or None to use the default options

        Returns:
            one alignment (:class:`~pysam.AlignedSegment`) per query
            :code:`List[List[AlignedSegment]]`.
        """
        if len(queries) == 0:
            return []
        elif isinstance(queries[0], str):
            queries = [
                FastxRecord(name=f"read.{i}", sequence=sequence)
                for i, sequence in enumerate(queries)
            ]
        opt = BwaAlnOptions() if opt is None else opt


        return self._calign(opt,  queries)
    
    @staticmethod
    def __to_str(_bytes: bytes) -> str:
        return _bytes.decode('utf-8')

    cdef _copy_seq(self, q: FastxRecord, bwa_seq_t* s, is_comp: bool):
        seq_len = len(q.sequence)
        s.len = seq_len
        s.clip_len = seq_len
        s.full_len = seq_len
        s.seq = <uint8_t *> calloc(sizeof(uint8_t), seq_len + 1)
        s.rseq = <uint8_t *> calloc(sizeof(uint8_t), seq_len + 1)
        s.qual = <uint8_t *> calloc(sizeof(uint8_t), seq_len + 1)

        # use seq_reverse from bwaseqio.c
        for i, base in enumerate(q.sequence):
            s.seq[i] = nst_nt4_table[ord(base)]
            s.rseq[i] = nst_nt4_table[ord(base)]
            s.qual[i] = 40  # FIXME: parameterize
        s.seq[seq_len] = b'\0'
        s.qual[seq_len] = b'\0'
        seq_reverse(seq_len, s.seq,
                    0)  #  // *IMPORTANT*: will be reversed back in bwa_refine_gapped()
        seq_reverse(seq_len, s.rseq, 1 if is_comp else 0)

        s.name = <char *> calloc(sizeof(char), len(q.name) + 1)
        strncpy(s.name, force_bytes(q.name), len(q.name))
        s.name[len(q.name)] = b'\0'

    cdef _build_alignment(self, query: FastxRecord, bwa_seq_t *seq, opt: BwaAlnOptions, kstring_t *kstr):
        cdef int reference_id
        cdef int nm
        cdef char XT
        cdef bwt_multi1_t* hit

        # make a default, unmapped, empty record
        rec = AlignedSegment(header=self._index.header)
        rec.query_name = query.name
        rec.reference_id = -1
        rec.reference_start = -1
        rec.mapping_quality = 0
        rec.is_paired = False
        rec.is_read1 = False
        rec.is_read2 = False
        rec.is_qcfail = False
        rec.is_duplicate = False
        rec.is_secondary = False
        rec.is_supplementary = False
        rec.is_unmapped = True
        rec.query_sequence = query.sequence
        rec.query_qualities = qualitystring_to_array(query.quality)

        if seq.type == BWA_TYPE_NO_MATCH:  # unmapped read
            # TODO: custom bwa tags: RG, BC, XC
            return rec

        # get the span of reference bases
        ref_len_in_alignment = pos_end(seq) - seq.pos

        # reference id
        nn = bns_cnt_ambi(self._index.bns(), seq.pos, ref_len_in_alignment, &reference_id)
        rec.reference_id = reference_id

        # make this unmapped if we map off the end of the contig
        rec.is_unmapped = seq.pos + ref_len_in_alignment - self._index.bns().anns[
            rec.reference_id].offset > self._index.bns().anns[rec.reference_id].len

        # strand, reference start, and mapping quality
        if seq.strand:
            rec.is_reverse = True
            rec.query_sequence = reverse_complement(query.sequence)
            if query.quality is not None:
                rec.query_qualities = qualitystring_to_array(query.quality[::-1])
        rec.reference_start = seq.pos - self._index.bns().anns[rec.reference_id].offset  # 0-based
        rec.mapping_quality = seq.mapQ

        # cigar
        cigar = ""
        if seq.cigar:
            for j in range(seq.n_cigar):
                cigar_len = __cigar_len(seq.cigar[j])
                cigar_op = "MIDS"[__cigar_op(seq.cigar[j])]
                cigar = f"{cigar}{cigar_len}{cigar_op}"
        elif seq.type != BWA_TYPE_NO_MATCH:
            cigar = f"{seq.len}M"
        rec.cigarstring = cigar

        # # tags
        if seq.type != BWA_TYPE_NO_MATCH:
            attrs = dict()
            tag = "NM" if ((opt._delegate.mode & BWA_MODE_COMPREAD) != 0) else "CM"
            attrs[tag] = int(seq.nm)
            if nn > 0:
                attrs["XN"] = nn
            if seq.type != BWA_TYPE_MATESW:  # X0 and X1 are not available for this type of alignment
                attrs["X0"] = seq.c1
                if seq.c1 <= opt.stop_at_max_best_hits:
                    attrs["X1"] = seq.c2
            # SM, AM, X0, and X1 are for paired end reads, so omitted
            attrs["XM"] = seq.n_mm
            attrs["XO"] = seq.n_gapo
            attrs["XG"] = seq.n_gapo + seq.n_gape
            if seq.md != NULL:
                attrs["MD"] = self.__to_str(seq.md)
            # print multiple hits
            if seq.n_multi > 0:
                XA = ""
                for j in range(seq.n_multi):
                    hit = &seq.multi[j]
                    end = pos_end_multi(hit, seq.len - hit.pos)
                    nn = bns_cnt_ambi(self._index.bns(), hit.pos, end, &reference_id)
                    XA += self.__to_str(self._index.bns().anns[reference_id].name)
                    XA += "," + ('-' if hit.strand != 0 else '+')
                    XA += str(hit.pos - self._index.bns().anns[reference_id].offset + 1)
                    XA += ","
                    if hit.cigar == NULL:
                        XA += f"{seq.len}M"
                    else:
                        for k in range(hit.n_cigar):
                            cigar_len = __cigar_len(hit.cigar[k])
                            cigar_op = "MIDS"[__cigar_op(hit.cigar[k])]
                            XA += f"{cigar_len}{cigar_op}"
                    XA += f",{hit.gap + hit.mm};"
                attrs["XA"] = XA
            # NB: the type 'A' (printable character) of the XA tag must be explicitly given
            # below, therefore we extract the attrs dictionary as a list of tuples first.
            tags = [('XT', b'N' if nn > 10 else "NURM"[seq.type], 'A')]
            tags.extend(list(attrs.items()))
            rec.set_tags(tags)

        return rec

    cdef _calign(self, opt: BwaAlnOptions, queries: List[FastxRecord]):
        cdef bwa_seq_t* seqs
        cdef bwa_seq_t* s
        cdef char* s_char
        cdef kstring_t* kstr
        cdef gap_opt_t* gap_opt

        gap_opt = opt.gap_opt()

        kstr = <kstring_t*>calloc(sizeof(kstring_t), 1)

        # copy FastqProxy into bwa_seq_t
        num_seqs = len(queries)
        seqs = <bwa_seq_t*>calloc(sizeof(bwa_seq_t), num_seqs)
        for i in range(num_seqs):
            self._copy_seq(queries[i], &seqs[i], (opt._delegate.mode & BWA_MODE_COMPREAD) != 0)
            seqs[i].tid = -1

        # this is `bwa aln`, and the rest is `bwa samse`
        bwa_cal_sa_reg_gap(0, self._index.bwt(), num_seqs, seqs, gap_opt)

        # create the full alignment
        for i in range(num_seqs):
            s = &seqs[i]
            # bwa_cal_sa_reg_gap frees name, seq, rseq, and qual, so add them back in again
            self._copy_seq(queries[i], s, (opt._delegate.mode & BWA_MODE_COMPREAD) != 0)
            bwa_aln2seq_core(s.n_aln, s.aln, s, 1, opt.max_hits)

        # # calculate the genomic position given the suffix array offsite
        bwa_cal_pac_pos_with_bwt(self._index.bns(), num_seqs, seqs, gap_opt.max_diff, gap_opt.fnr, self._index.bwt())

        # refine gapped alignment
        bwa_refine_gapped(self._index.bns(), num_seqs, seqs, self._index.pac())

        # create the AlignedSegment from FastxRecord and bwa_seq_t.
        recs = [
            self._build_alignment(query=queries[i], seq=&seqs[i], opt=opt, kstr=kstr)
            for i in range(num_seqs)
        ]

        bwa_free_read_seq(num_seqs, seqs)
        free(kstr.s)
        free(kstr)

        return recs



# cython: language_level=3

from libc.stdint cimport uint8_t, int64_t, int32_t, uint64_t, int8_t, uint32_t
from libc.stdio cimport FILE

cdef extern from "limits.h":
    cdef int INT_MAX

cdef extern from "bwt.h":
    ctypedef struct bwt_t:
        int sa_intv


cdef extern from "bntseq.h":
    ctypedef  struct bntann1_t:
        int64_t offset
        int32_t len
        char *name

    ctypedef struct bntseq_t:
        int64_t l_pac
        bntann1_t *anns
        FILE * fp_pac

    unsigned char nst_nt4_table[256]

cdef extern from "kstring.h":
    ctypedef struct kstring_t:
        size_t l, m
        char *s


cdef extern from "bwamem.h":
    int MEM_F_PE        
    int MEM_F_NOPAIRING 
    int MEM_F_ALL       
    int MEM_F_NO_MULTI  
    int MEM_F_NO_RESCUE 
    int MEM_F_REF_HDR   
    int MEM_F_SOFTCLIP  
    int MEM_F_SMARTPE   
    int MEM_F_PRIMARY5  
    int MEM_F_KEEP_SUPP_MAPQ 
    int MEM_F_XB

    ctypedef struct mem_pestat_t:
        int low, high   # lower and upper bounds within which a read pair is considered to be properly paired
        int failed      # non-zero if the orientation is not supported by sufficient data
        double avg, std # mean and stddev of the insert size distribution

    ctypedef struct mem_opt_t:
        int a, b                # match score and mismatch penalty
        int o_del, e_del
        int o_ins, e_ins
        int pen_unpaired        # phred-scaled penalty for unpaired reads
        int pen_clip5,pen_clip3 # clipping penalty. This score is not deducted from the DP score.
        int w                   # band width
        int zdrop               # Z-dropoff
        uint64_t max_mem_intv
        int T                  # output score threshold only affecting output
        int flag               # see MEM_F_* macros
        int min_seed_len       # minimum seed length
        int min_chain_weight
        int max_chain_extend
        float split_factor     # split into a seed if MEM is longer than min_seed_len*split_factor
        int split_width        # split into a seed if its occurence is smaller than this value
        int max_occ            # skip a seed if its occurence is larger than this value
        int max_chain_gap      # do not chain seed if it is max_chain_gap-bp away from the closest seed
        int n_threads          # number of threads
        int chunk_size         # process chunk_size-bp sequences in a batch
        float mask_level       # regard a hit as redundant if the overlap with another better hit is over mask_level times the min length of the two hits
        float drop_ratio       # drop a chain if its seed coverage is below drop_ratio times the seed coverage of a better chain overlapping with the small chain
        float XA_drop_ratio    # when counting hits for the XA tag, ignore alignments with score < XA_drop_ratio * max_score only effective for the XA tag
        float mask_level_redun
        float mapQ_coef_len
        int mapQ_coef_fac
        int max_ins            # when estimating insert size distribution, skip pairs with insert longer than this value
        int max_matesw         # perform maximally max_matesw rounds of mate-SW for each end
        int max_XA_hits, max_XA_hits_alt # if there are max_hits or fewer, output them all
        int8_t mat[25]         # scoring matrix mat[0] == 0 if unset
    ctypedef struct mem_alnreg_t:
        int score  # best local SW score
        int secondary  # index of the parent hit shadowing the current hit; <0 if primary
        int n_comp  # number of sub-alignments chained together
        int is_alt

    ctypedef struct mem_alnreg_v:
        size_t n, m
        mem_alnreg_t *a

    ctypedef struct mem_aln_t:
        int64_t pos     # forward strand 5'-end mapping position
        int rid         # reference sequence index in bntseq_t; <0 for unmapped
        int flag        # extra flag
        uint32_t is_rev  # is_rev: whether on the reverse strand;
        uint32_t is_alt
        uint32_t mapq   # mapq: mapping quality;
        uint32_t NM  # NM: edit distance
        int n_cigar;     # number of CIGAR operations
        uint32_t *cigar; # CIGAR in the BAM encoding: opLen<<4|op; op to integer mapping: MIDSH=>01234
        char *XA
        int score, sub, alt_sc;

    mem_opt_t *mem_opt_init()
    mem_alnreg_v mem_align1(const mem_opt_t *opt, const bwt_t *bwt, const bntseq_t *bns, const uint8_t *pac, int l_seq, const char *seq)
    void bwa_fill_scmat(int a, int b, int8_t mat[25])
    mem_aln_t mem_reg2aln(const mem_opt_t *opt, const bntseq_t *bns, const uint8_t *pac, int l_seq, const char *seq, const mem_alnreg_t *ar)

# from bwamem.c
cdef extern void mem_reorder_primary5(int T, mem_alnreg_v *a)
cdef extern void add_cigar(const mem_opt_t *opt, mem_aln_t *p, kstring_t *str, int which)

# from bwamem_extra.c
cdef extern char **mem_gen_alt(const mem_opt_t *opt, const bntseq_t *bns, const uint8_t *pac, mem_alnreg_v *a, int l_query, const char *query);





# cython: language_level=3

from libc.stdint cimport uint8_t, uint64_t, uint16_t, uint32_t, int64_t, int32_t
from libc.stdio cimport FILE

cdef extern from "libbwaaln_utils.h":
    void bwa_cal_pac_pos_with_bwt(const bntseq_t *bns, int n_seqs, bwa_seq_t *seqs, int max_mm,
                                  float fnr, bwt_t *bwt)

cdef extern from "bntseq.h":
    unsigned char nst_nt4_table[256]
    int bns_cnt_ambi(const bntseq_t *bns, int64_t pos_f, int len, int *ref_id)

cdef extern from "bwt.h":
    ctypedef struct bwt_t:
        int sa_intv

cdef extern from "bwtaln.h":
    int BWA_TYPE_NO_MATCH
    int BWA_TYPE_MATESW

    int BWA_MODE_GAPE
    int BWA_MODE_COMPREAD
    int BWA_MODE_LOGGAP

    int __cigar_op(uint16_t __cigar)
    int __cigar_len(uint16_t __cigar)

    ctypedef struct gap_opt_t:
        int trim_qual
        int s_mm
        int s_gapo
        int s_gape
        int mode # bit 24-31 are the barcode length
        int indel_end_skip
        int max_del_occ
        int max_entries
        float fnr
        int max_diff
        int max_gapo
        int max_gape
        int max_seed_diff
        int seed_len
        int n_threads
        int max_top2
        int trim_qual
        int sam
        char *rg_line
        int n_occ
        int interactive_mode
        int with_md

    gap_opt_t *gap_init_opt()
    void gap_print_opt(const gap_opt_t *opt)

    void seq_reverse(int len, unsigned char *seq, int is_comp)

    ctypedef struct bwt_aln1_t:
        pass

cdef extern from "bntseq.h":
    ctypedef  struct bntann1_t:
        int64_t offset
        int32_t len
        char *name

    ctypedef struct bntseq_t:
        int64_t l_pac
        bntann1_t *anns
        FILE * fp_pac

    bntseq_t * bns_restore(const char * prefix)
    void bns_destroy(bntseq_t *bns)

cdef extern from "kseq.h":
    ctypedef struct kstring_t:
        char *s

cdef extern void bwa_aln2seq_core(int n_aln, const bwt_aln1_t *aln, bwa_seq_t *s, int set_main, int n_multi)

cdef extern from "bwase.h":
    int64_t pos_end(const bwa_seq_t *p)
    void bwa_refine_gapped(const bntseq_t *bns, int n_seqs, bwa_seq_t *seqs, unsigned char *_pacseq)
    char *bwa_cal_md1(int n_cigar, uint16_t *cigar, int len, uint64_t pos, unsigned char *seq, uint64_t l_pac, unsigned char *pacseq, kstring_t *str, int *_nm)
    void bwase_initialize()

cdef extern from "bwtaln.h":
    ctypedef struct bwt_multi1_t:
        uint32_t n_cigar
        uint32_t gap
        uint32_t mm
        uint32_t strand
        int ref_shift
        uint64_t pos
        uint16_t *cigar
    ctypedef struct bwa_seq_t:
        char *name
        uint8_t *seq
        uint8_t *rseq
        uint8_t *qual
        uint32_t len
        uint32_t strand
        uint32_t type
        uint32_t n_mm
        uint32_t n_gapo
        uint32_t n_gape
        uint32_t mapQ
        int score
        int clip_len
        int n_aln
        bwt_aln1_t *aln
        int n_multi;
        bwt_multi1_t *multi;
        uint64_t pos
        uint64_t c1
        uint64_t c2
        uint16_t *cigar
        int n_cigar
        int tid
        uint32_t full_len
        uint32_t nm
        char *md


    void bwa_free_read_seq(int n_seqs, bwa_seq_t *seqs)

    void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt, int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt)


# from bwase.c
cdef extern int64_t pos_end_multi(const bwt_multi1_t *p, int len)



import enum
from pathlib import Path

from pysam import AlignmentHeader

ERROR_HANDLER: str
TEXT_ENCODING: str

@enum.unique
class BwaIndexBuildMethod(enum.Enum):
    """The BWT construction algorithm (:code:`bwa index -a <str>`)"""

    AUTO = enum.auto()
    """Choose the algorithm automatically (based on the reference length)"""
    RB2 = enum.auto()
    """rb2 (does not work for "long" genomes)"""
    BWTSW = enum.auto()
    """bwtsw (does not work for "short" genomes)"""
    IS = enum.auto()
    """is (does not work for "long" genomes)"""

class BwaIndex:
    header: AlignmentHeader
    def __init__(
        self, prefix: str | Path, bwt: bool = ..., bns: bool = ..., pac: bool = ...
    ) -> None: ...
    @classmethod
    def index(
        cls,
        fasta: str | Path,
        method: BwaIndexBuildMethod = BwaIndexBuildMethod.AUTO,
        prefix: str | Path | None = None,
        block_size: int = 10000000,
        out_64: bool = False,
    ) -> None: ...



import enum
from pathlib import Path
from typing import List

from pysam import AlignedSegment
from pysam import FastxRecord

from pybwa.libbwaindex import BwaIndex

# class syntax
@enum.unique
class BwaMemMode(enum.Enum):
    """The read type for overriding multiple options"""

    PACBIO = enum.auto()
    ONT2D = enum.auto()
    INTRACTG = enum.auto()

class BwaMemOptions:
    def __init__(
        self,
        min_seed_len: int | None = None,
        mode: BwaMemMode | None = None,
        band_width: int | None = None,
        match_score: int | None = None,
        mismatch_penalty: int | None = None,
        minimum_score: int | None = None,
        unpaired_penalty: int | None = None,
        n_threads: int | None = None,
        skip_pairing: bool | None = None,
        output_all_for_fragments: bool | None = None,
        interleaved_paired_end: bool | None = None,
        short_split_as_secondary: bool | None = None,
        skip_mate_rescue: bool | None = None,
        soft_clip_supplementary: bool | None = None,
        with_xr_tag: bool | None = None,
        query_coord_as_primary: bool | None = None,
        keep_mapq_for_supplementary: bool | None = None,
        with_xb_tag: bool | None = None,
        max_occurrences: int | None = None,
        off_diagonal_x_dropoff: float | None = None,
        ignore_alternate_contigs: bool | None = None,
        internal_seed_split_factor: float | None = None,
        drop_chain_fraction: float | None = None,
        max_mate_rescue_rounds: int | None = None,
        min_seeded_bases_in_chain: int | None = None,
        seed_occurrence_in_3rd_round: int | None = None,
        xa_max_hits: int | tuple[int, int] | None = None,
        xa_drop_ratio: float | None = None,
        gap_open_penalty: int | tuple[int, int] | None = None,
        gap_extension_penalty: int | tuple[int, int] | None = None,
        clipping_penalty: int | tuple[int, int] | None = None,
    ) -> None: ...
    _finalized: bool
    _ignore_alt: bool
    _mode: BwaMemMode | None
    @property
    def finalized(self) -> bool: ...
    @property
    def min_seed_len(self) -> int: ...
    @min_seed_len.setter
    def min_seed_len(self, value: int) -> None: ...
    @property
    def mode(self) -> BwaMemMode: ...
    @mode.setter
    def mode(self, value: BwaMemMode) -> None: ...
    @property
    def band_width(self) -> int: ...
    @band_width.setter
    def band_width(self, value: int) -> None: ...
    @property
    def match_score(self) -> int: ...
    @match_score.setter
    def match_score(self, value: int) -> None: ...
    @property
    def mismatch_penalty(self) -> int: ...
    @mismatch_penalty.setter
    def mismatch_penalty(self, value: int) -> None: ...
    @property
    def minimum_score(self) -> int: ...
    @minimum_score.setter
    def minimum_score(self, value: int) -> None: ...
    @property
    def unpaired_penalty(self) -> int: ...
    @unpaired_penalty.setter
    def unpaired_penalty(self, value: int) -> None: ...
    @property
    def n_threads(self) -> int: ...
    @n_threads.setter
    def n_threads(self, value: int) -> None: ...
    @property
    def skip_pairing(self) -> bool: ...
    @skip_pairing.setter
    def skip_pairing(self, value: bool) -> None: ...
    @property
    def output_all_for_fragments(self) -> bool: ...
    @output_all_for_fragments.setter
    def output_all_for_fragments(self, value: bool) -> None: ...
    @property
    def interleaved_paired_end(self) -> bool: ...
    @interleaved_paired_end.setter
    def interleaved_paired_end(self, value: bool) -> None: ...
    @property
    def short_split_as_secondary(self) -> bool: ...
    @short_split_as_secondary.setter
    def short_split_as_secondary(self, value: bool) -> None: ...
    @property
    def skip_mate_rescue(self) -> bool: ...
    @skip_mate_rescue.setter
    def skip_mate_rescue(self, value: bool) -> None: ...
    @property
    def soft_clip_supplementary(self) -> bool: ...
    @soft_clip_supplementary.setter
    def soft_clip_supplementary(self, value: bool) -> None: ...
    @property
    def with_xr_tag(self) -> bool: ...
    @with_xr_tag.setter
    def with_xr_tag(self, value: bool) -> None: ...
    @property
    def query_coord_as_primary(self) -> bool: ...
    @query_coord_as_primary.setter
    def query_coord_as_primary(self, value: bool) -> None: ...
    @property
    def keep_mapq_for_supplementary(self) -> bool: ...
    @keep_mapq_for_supplementary.setter
    def keep_mapq_for_supplementary(self, value: bool) -> None: ...
    @property
    def with_xb_tag(self) -> bool: ...
    @with_xb_tag.setter
    def with_xb_tag(self, value: bool) -> None: ...
    @property
    def max_occurrences(self) -> int: ...
    @max_occurrences.setter
    def max_occurrences(self, value: int) -> None: ...
    @property
    def off_diagonal_x_dropoff(self) -> float: ...
    @off_diagonal_x_dropoff.setter
    def off_diagonal_x_dropoff(self, value: float) -> None: ...
    @property
    def ignore_alternate_contigs(self) -> bool: ...
    @ignore_alternate_contigs.setter
    def ignore_alternate_contigs(self, value: bool) -> None: ...
    @property
    def internal_seed_split_factor(self) -> float: ...
    @internal_seed_split_factor.setter
    def internal_seed_split_factor(self, value: float) -> None: ...
    @property
    def drop_chain_fraction(self) -> float: ...
    @drop_chain_fraction.setter
    def drop_chain_fraction(self, value: float) -> None: ...
    @property
    def max_mate_rescue_rounds(self) -> int: ...
    @max_mate_rescue_rounds.setter
    def max_mate_rescue_rounds(self, value: int) -> None: ...
    @property
    def min_seeded_bases_in_chain(self) -> int: ...
    @min_seeded_bases_in_chain.setter
    def min_seeded_bases_in_chain(self, value: int) -> None: ...
    @property
    def seed_occurrence_in_3rd_round(self) -> int: ...
    @seed_occurrence_in_3rd_round.setter
    def seed_occurrence_in_3rd_round(self, value: int) -> None: ...
    @property
    def xa_max_hits(self) -> int | tuple[int, int]: ...
    @xa_max_hits.setter
    def xa_max_hits(self, value: int | tuple[int, int]) -> None: ...
    @property
    def xa_drop_ratio(self) -> float: ...
    @xa_drop_ratio.setter
    def xa_drop_ratio(self, value: float) -> None: ...
    @property
    def gap_open_penalty(self) -> int | tuple[int, int]: ...
    @gap_open_penalty.setter
    def gap_open_penalty(self, value: int | tuple[int, int]) -> None: ...
    @property
    def gap_extension_penalty(self) -> int | tuple[int, int]: ...
    @gap_extension_penalty.setter
    def gap_extension_penalty(self, value: int | tuple[int, int]) -> None: ...
    @property
    def clipping_penalty(self) -> int | tuple[int, int]: ...
    @clipping_penalty.setter
    def clipping_penalty(self, value: int | tuple[int, int]) -> None: ...
    def finalize(self, copy: bool = False) -> BwaMemOptions: ...

class BwaMem:
    _index: BwaIndex
    def __init__(self, prefix: str | Path | None = None, index: BwaIndex | None = None) -> None: ...
    def align(
        self, queries: List[FastxRecord] | List[str], opt: BwaMemOptions | None = None
    ) -> List[List[AlignedSegment]]: ...



# cython: language_level=3
from pathlib import Path
from typing import List
from fgpyo.sequence import reverse_complement
from libc.string cimport memset, memcpy
from libc.stdlib cimport calloc, free
import enum
from pybwa.libbwaindex cimport BwaIndex
from pysam import FastxRecord, AlignedSegment, qualitystring_to_array


__all__ = [
    "BwaMemMode",
    "BwaMemOptions",
    "BwaMem",
]


@enum.unique
class BwaMemMode(enum.Enum):
    """The read type for overriding multiple options"""
    PACBIO = enum.auto()
    """PacBio reads to ref"""
    ONT2D = enum.auto()
    """Oxford Nanopore 2D-reads to ref"""
    INTRACTG = enum.auto()
    """intra-species contigs to ref"""


cdef class BwaMemOptions:
    """The container for options for :class:`~pybwa.BwaMem`.

    Args:
        min_seed_len (int): :code:`bwa mem -k <int>`
        mode (BwaMemMode): :code:`bwa mem -x <str>`
        band_width (int): :code:`bwa mem -w <int>`
        match_score (int): :code:`bwa mem -A <int>`
        mismatch_penalty (int): :code:`bwa mem -A <int>`
        minimum_score (int): :code:`bwa mem -T <int>`
        unpaired_penalty (int): :code:`bwa mem -U <int>`
        n_threads (int): :code:`bwa mem -t <int>`
        skip_pairing (bool): :code:`bwa mem -P`
        output_all_for_fragments (bool): :code:`bwa mem -a`
        interleaved_paired_end (bool): :code:`bwa mem -p`
        short_split_as_secondary (bool): :code:`bwa mem -M`
        skip_mate_rescue (bool): :code:`bwa mem -S`
        soft_clip_supplementary (bool): :code:`bwa mem -Y`
        with_xr_tag (bool): :code:`bwa mem -V`
        query_coord_as_primary (bool): :code:`bwa mem -5`
        keep_mapq_for_supplementary (bool): :code:`bwa mem -q`
        with_xb_tag (bool): :code:`bwa mem -u`
        max_occurrences (int): :code:`bwa mem -c <int>`
        off_diagonal_x_dropoff (float): :code:`bwa mem -d <float>`
        ignore_alternate_contigs (bool): :code:`bwa mem -j`
        internal_seed_split_factor (float): :code:`bwa mem -r <float>`
        drop_chain_fraction (float): :code:`bwa mem -D <float>`
        max_mate_rescue_rounds (int): :code:`bwa mem -m <int>`
        min_seeded_bases_in_chain (int): :code:`bwa mem -W <int>`
        seed_occurrence_in_3rd_round (int): :code:`bwa mem -y <int>`
        xa_max_hits (int | tuple[int, int]): :code:`bwa mem -h <int<,int>>`
        xa_drop_ratio (float): :code:`bwa mem -y <float>`
        gap_open_penalty (int | tuple[int, int]): :code:`bwa mem -O <int<,int>>`
        gap_extension_penalty (int | tuple[int, int]): :code:`bwa mem -E <int<,int>>`
        clipping_penalty (int | tuple[int, int]): :code:`bwa mem -L <int<,int>>`
    """
    _finalized: bool
    _ignore_alt: bool
    _mode: BwaMemMode | None
    cdef mem_opt_t* _options
    cdef mem_opt_t* _options0

    def _assert_not_finalized(self, attr_name: str) -> None:
        """Raises an AttributeError if the options have been finalized.

        This is used in each setter below to enforce that setters are not used after
         :meth:`~pybwa.BwaMemOptions.finalize` is called.  This could be a property decorator,
         but cython does not support decorating properties.
        """
        if self._finalized:
            raise AttributeError(f"can't set attribute: {attr_name}")

    def __init__(self,
                 min_seed_len: int | None = None,
                 mode: BwaMemMode | None = None,
                 band_width: int | None = None,
                 match_score: int | None = None,
                 mismatch_penalty: int | None = None,
                 minimum_score: int | None = None,
                 unpaired_penalty: int | None = None,
                 n_threads: int | None = None,
                 skip_pairing: bool | None = None,
                 output_all_for_fragments: bool | None = None,
                 interleaved_paired_end: bool | None = None,
                 short_split_as_secondary: bool | None = None,
                 skip_mate_rescue: bool | None = None,
                 soft_clip_supplementary: bool | None = None,
                 with_xr_tag: bool | None = None,
                 query_coord_as_primary: bool | None = None,
                 keep_mapq_for_supplementary: bool | None = None,
                 with_xb_tag: bool | None = None,
                 max_occurrences: int | None = None,
                 off_diagonal_x_dropoff: float | None = None,
                 ignore_alternate_contigs: bool | None = None,
                 internal_seed_split_factor: float | None = None,
                 drop_chain_fraction: float | None = None,
                 max_mate_rescue_rounds: int | None = None,
                 min_seeded_bases_in_chain: int | None = None,
                 seed_occurrence_in_3rd_round: int | None = None,
                 xa_max_hits: int | tuple[int, int] | None = None,
                 xa_drop_ratio: float | None = None,
                 gap_open_penalty: int | tuple[int, int] | None = None,
                 gap_extension_penalty: int | tuple[int, int] | None = None,
                 clipping_penalty: int | tuple[int, int] | None = None) -> None:
        self._finalized = False
        self._ignore_alt = False
        self._mode = None
        if min_seed_len is not None:
            self.min_seed_len = min_seed_len
        if mode is not None:
            self.mode = mode
        if band_width is not None:
            self.band_width = band_width
        if match_score is not None:
            self.match_score = match_score
        if mismatch_penalty is not None:
            self.mismatch_penalty = mismatch_penalty
        if minimum_score is not None:
            self.minimum_score = minimum_score
        if unpaired_penalty is not None:
            self.unpaired_penalty = unpaired_penalty
        if n_threads is not None:
            self.n_threads = n_threads
        if skip_pairing is not None:
            self.skip_pairing = skip_pairing
        if output_all_for_fragments is not None:
            self.output_all_for_fragments = output_all_for_fragments
        if interleaved_paired_end is not None:
            self.interleaved_paired_end = interleaved_paired_end
        if short_split_as_secondary is not None:
            self.short_split_as_secondary = short_split_as_secondary
        if skip_mate_rescue is not None:
            self.skip_mate_rescue = skip_mate_rescue
        if soft_clip_supplementary is not None:
            self.soft_clip_supplementary = soft_clip_supplementary
        if with_xr_tag is not None:
            self.with_xr_tag = with_xr_tag
        if query_coord_as_primary is not None:
            self.query_coord_as_primary = query_coord_as_primary
        if keep_mapq_for_supplementary is not None:
            self.keep_mapq_for_supplementary = keep_mapq_for_supplementary
        if with_xb_tag is not None:
            self.with_xb_tag = with_xb_tag
        if max_occurrences is not None:
            self.max_occurrences = max_occurrences
        if off_diagonal_x_dropoff is not None:
            self.off_diagonal_x_dropoff = off_diagonal_x_dropoff
        if ignore_alternate_contigs is not None:
            self.ignore_alternate_contigs = ignore_alternate_contigs
        if internal_seed_split_factor is not None:
            self.internal_seed_split_factor = internal_seed_split_factor
        if drop_chain_fraction is not None:
            self.drop_chain_fraction = drop_chain_fraction
        if max_mate_rescue_rounds is not None:
            self.max_mate_rescue_rounds = max_mate_rescue_rounds
        if min_seeded_bases_in_chain is not None:
            self.min_seeded_bases_in_chain = min_seeded_bases_in_chain
        if seed_occurrence_in_3rd_round is not None:
            self.seed_occurrence_in_3rd_round = seed_occurrence_in_3rd_round
        if xa_max_hits is not None:
            self.xa_max_hits = xa_max_hits
        if xa_drop_ratio is not None:
            self.xa_drop_ratio = xa_drop_ratio
        if gap_open_penalty is not None:
            self.gap_open_penalty = gap_open_penalty
        if gap_extension_penalty is not None:
            self.gap_extension_penalty = gap_extension_penalty
        if clipping_penalty is not None:
            self.clipping_penalty = clipping_penalty
        
    cdef _copy_options(self, dst: BwaMemOptions, src: BwaMemOptions):
        memcpy(dst._options, src._options, sizeof(mem_opt_t))
        memcpy(dst._options0, src._options0, sizeof(mem_opt_t))

    def __cinit__(self, opt: BwaMemOptions | None = None):
        self._options = mem_opt_init()
        self._options0 = mem_opt_init()
        if opt is not None:
            self._copy_options(self, opt)

    def __dealloc__(self):
        free(self._options)
        self._options = NULL
        free(self._options0)
        self._options0 = NULL

    cdef mem_opt_t* mem_opt(self):
        """Returns the options struct to use with the bwa C library methods"""
        if not self._finalized:
            raise Exception("Cannot call `mem_opt` until `finalize()` is called")
        return self._options

    @property
    def finalized(self) -> bool:
        """True if the options have been finalized with :meth:`~pybwa.BwaMemOptions.finalize`."""
        return self._finalized

    def finalize(self, copy: bool = False) -> BwaMemOptions:
        """Performs final initialization of these options.  The object returned may not be
        modified further.

        If the mode is given, then the presets are applied to options that have not been explicitly
        set.  Otherwise, if the match score has been set, the match score scales the (-TdBOELU)
        options if they have not been explicitly set.

        Args:
            copy: true to return a finalized copy of this object, false to finalize this object
        """
        opt: BwaMemOptions
        if copy:
            opt = BwaMemOptions()
            memcpy(opt._options, self._options, sizeof(mem_opt_t))
        else:
            opt = self

        if opt._mode is None:
            # matching score is changed so scale the rest of the penalties/scores
            if opt._options0.a == 1:
                if opt._options0.b != 1:
                    opt._options.b *= opt._options.a
                if opt._options0.T != 1:
                    opt._options.T *= opt._options.a
                if opt._options0.o_del != 1:
                    opt._options.o_del *= opt._options.a
                if opt._options0.e_del != 1:
                    opt._options.e_del *= opt._options.a
                if opt._options0.o_ins != 1:
                    opt._options.o_ins *= opt._options.a
                if opt._options0.e_ins != 1:
                    opt._options.e_ins *= opt._options.a
                if opt._options0.zdrop != 1:
                    opt._options.zdrop *= opt._options.a
                if opt._options0.pen_clip5 != 1:
                    opt._options.pen_clip5 *= opt._options.a
                if opt._options0.pen_clip3 != 1:
                    opt._options.pen_clip3 *= opt._options.a
                if opt._options0.pen_unpaired != 1:
                    opt._options.pen_unpaired *= opt._options.a
        elif opt._mode == BwaMemMode.INTRACTG:
            if opt._options0.o_del != 1:
                opt._options.o_del = 16
            if opt._options0.o_ins != 1:
                opt._options.o_ins = 16
            if opt._options0.b != 1:
                opt._options.b = 9
            if opt._options0.pen_clip5 != 1:
                opt._options.pen_clip5 = 5
            if opt._options0.pen_clip3 != 1:
                opt._options.pen_clip3 = 5
        else:
            if opt._options0.o_del != 1:
                opt._options.o_del = 1
            if opt._options0.o_ins != 1:
                opt._options.o_ins = 1
            if opt._options0.e_del != 1:
                opt._options.e_del = 1
            if opt._options0.e_ins != 1:
                opt._options.e_ins = 1
            if opt._options0.b != 1:
                opt._options.b = 1
            if opt._options0.split_factor == 0.0:
                opt._options.split_factor = 10.0
            if opt._options0.pen_clip5 != 1:
                opt._options.pen_clip5 = 0
            if opt._options0.pen_clip3 != 1:
                opt._options.pen_clip3 = 0
            # ONT2D vs PACBIO options
            if opt._options0.min_chain_weight != 1:
                opt._options.min_chain_weight = 20 if opt._mode == BwaMemMode.ONT2D else 40
            if opt._options0.min_seed_len != 1:
                opt._options.min_seed_len = 14 if opt._mode == BwaMemMode.ONT2D else 17

        bwa_fill_scmat(
            opt._options.a, opt._options.b, opt._options.mat
        )

        opt._finalized = True
        return opt

    @property
    def min_seed_len(self) -> int:
        """:code:`bwa mem -k <int>`"""
        return self._options.min_seed_len

    @min_seed_len.setter
    def min_seed_len(self, value: int):
        self._assert_not_finalized(attr_name=BwaMemOptions.min_seed_len.__name__)
        self._options.min_seed_len = value
        self._options0.min_seed_len = 1

    @property
    def mode(self) -> BwaMemMode:
        """:code:`bwa mem -x <str>`"""
        return self._mode

    @mode.setter
    def mode(self, value: BwaMemMode) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.mode.__name__)
        self._mode = value

    @property
    def band_width(self) -> int:
        """:code:`bwa mem -w <int>`"""
        return self._options.w

    @band_width.setter
    def band_width(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.band_width.__name__)
        self._options.w = value
        self._options0.w = 1

    @property
    def match_score(self) -> int:
        """:code:`bwa mem -A <int>`"""
        return self._options.a


    @match_score.setter
    def match_score(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.match_score.__name__)
        self._options.a = value
        self._options0.a = 1

    @property
    def mismatch_penalty(self) -> int:
        """:code:`bwa mem -A <int>`"""
        return self._options.b

    @mismatch_penalty.setter
    def mismatch_penalty(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.mismatch_penalty.__name__)
        self._options.b = value
        self._options0.b = 1

    @property
    def minimum_score(self) -> int:
        """:code:`bwa mem -T <int>`"""
        return self._options.T

    @minimum_score.setter
    def minimum_score(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.minimum_score.__name__)
        self._options.T = value
        self._options0.T = 1

    @property
    def unpaired_penalty(self) -> int:
        """:code:`bwa mem -U <int>`"""
        return self._options.pen_unpaired

    @unpaired_penalty.setter
    def unpaired_penalty(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.unpaired_penalty.__name__)
        self._options.pen_unpaired = value
        self._options0.pen_unpaired = 1

    @property
    def n_threads(self) -> int:
        """:code:`bwa mem -t <int>`"""
        return self._options.n_threads

    @n_threads.setter
    def n_threads(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.n_threads.__name__)
        self._options.n_threads = value if value > 1 else 1

    def _set_flag(self, value: bool, flag: int):
        if value:
            self._options.flag |= flag
        else:
            self._options.flag |= ~flag
        return self

    @property
    def skip_pairing(self) -> bool:
        """:code:`bwa mem -P`"""
        return (self._options.flag & MEM_F_NOPAIRING) != 0

    @skip_pairing.setter
    def skip_pairing(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.skip_pairing.__name__)
        self._set_flag(value, MEM_F_NOPAIRING)

    @property
    def output_all_for_fragments(self) -> bool:
        """:code:`bwa mem -a`"""
        return (self._options.flag & MEM_F_ALL) != 0

    @output_all_for_fragments.setter
    def output_all_for_fragments(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.output_all_for_fragments.__name__)
        self._set_flag(value, MEM_F_ALL)

    @property
    def interleaved_paired_end(self) -> bool:
        """:code:`bwa mem -p`"""
        return (self._options.flag & (MEM_F_PE | MEM_F_SMARTPE)) != 0

    @interleaved_paired_end.setter
    def interleaved_paired_end(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.interleaved_paired_end.__name__)
        self._set_flag(value, MEM_F_PE | MEM_F_SMARTPE)
    
    @property
    def short_split_as_secondary(self) -> bool:
        """:code:`bwa mem -M`"""
        return (self._options.flag & MEM_F_NO_MULTI) != 0

    @short_split_as_secondary.setter
    def short_split_as_secondary(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.short_split_as_secondary.__name__)
        self._set_flag(value, MEM_F_NO_MULTI)

    @property
    def skip_mate_rescue(self) -> bool:
        """:code:`bwa mem -S`"""
        return (self._options.flag & MEM_F_NO_RESCUE) != 0

    @skip_mate_rescue.setter
    def skip_mate_rescue(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.skip_mate_rescue.__name__)
        self._set_flag(value, MEM_F_NO_RESCUE)

    @property
    def soft_clip_supplementary(self) -> bool:
        """:code:`bwa mem -Y`"""
        return (self._options.flag & MEM_F_SOFTCLIP) != 0

    @soft_clip_supplementary.setter
    def soft_clip_supplementary(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.soft_clip_supplementary.__name__)
        self._set_flag(value, MEM_F_SOFTCLIP)

    @property
    def with_xr_tag(self) -> bool:
        """:code:`bwa mem -V`"""
        return (self._options.flag & MEM_F_REF_HDR) != 0

    @with_xr_tag.setter
    def with_xr_tag(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.with_xr_tag.__name__)
        self._set_flag(value, MEM_F_REF_HDR)

    @property
    def query_coord_as_primary(self) -> bool:
        """:code:`bwa mem -5`"""
        return (self._options.flag & (MEM_F_PRIMARY5 | MEM_F_KEEP_SUPP_MAPQ)) != 0

    @query_coord_as_primary.setter
    def query_coord_as_primary(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.query_coord_as_primary.__name__)
        self._set_flag(value, MEM_F_PRIMARY5 | MEM_F_KEEP_SUPP_MAPQ) # always apply MEM_F_KEEP_SUPP_MAPQ with -5

    @property
    def keep_mapq_for_supplementary(self) -> bool:
        """:code:`bwa mem -q`"""
        return (self._options.flag & MEM_F_KEEP_SUPP_MAPQ) != 0

    @keep_mapq_for_supplementary.setter
    def keep_mapq_for_supplementary(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.keep_mapq_for_supplementary.__name__)
        self._set_flag(value, MEM_F_KEEP_SUPP_MAPQ)

    @property
    def with_xb_tag(self) -> bool:
        """:code:`bwa mem -u`"""
        return (self._options.flag & MEM_F_XB) != 0

    @with_xb_tag.setter
    def with_xb_tag(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.with_xb_tag.__name__)
        self._set_flag(value, MEM_F_XB)

    @property
    def max_occurrences(self) -> int:
        """:code:`bwa mem -c <int>`"""
        return self._options.max_occ

    @max_occurrences.setter
    def max_occurrences(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.max_occurrences.__name__)
        self._options.max_occ = value
        self._options0.max_occ = 1

    @property
    def off_diagonal_x_dropoff(self) -> float:
        """:code:`bwa mem -d <float>`"""
        return self._options.XA_drop_ratio

    @off_diagonal_x_dropoff.setter
    def off_diagonal_x_dropoff(self, value: float) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.off_diagonal_x_dropoff.__name__)
        self._options.XA_drop_ratio = value
        self._options0.XA_drop_ratio = 1

    @property
    def ignore_alternate_contigs(self) -> bool:
        """:code:`bwa mem -j`"""
        return self._ignore_alt

    @ignore_alternate_contigs.setter
    def ignore_alternate_contigs(self, value: bool) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.ignore_alternate_contigs.__name__)
        self._ignore_alt = value

    @property
    def internal_seed_split_factor(self) -> float:
        """:code:`bwa mem -r <float>`"""
        return self._options.split_factor

    @internal_seed_split_factor.setter
    def internal_seed_split_factor(self, value: float) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.internal_seed_split_factor.__name__)
        self._options.split_factor = value
        self._options0.split_factor = 1

    @property
    def drop_chain_fraction(self) -> float:
        """:code:`bwa mem -D <float>`"""
        return self._options.drop_ratio

    @drop_chain_fraction.setter
    def drop_chain_fraction(self, value: float) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.drop_chain_fraction.__name__)
        self._options.drop_ratio = value
        self._options0.drop_ratio = 1

    @property
    def max_mate_rescue_rounds(self) -> int:
        """:code:`bwa mem -m <int>`"""
        return self._options.max_matesw

    @max_mate_rescue_rounds.setter
    def max_mate_rescue_rounds(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.max_mate_rescue_rounds.__name__)
        self._options.max_matesw = value
        self._options0.max_matesw = 1

    @property
    def min_seeded_bases_in_chain(self) -> int:
        """:code:`bwa mem -W <int>`"""
        return self._options.min_chain_weight

    @min_seeded_bases_in_chain.setter
    def min_seeded_bases_in_chain(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.min_seeded_bases_in_chain.__name__)
        self._options.min_chain_weight = value
        self._options0.min_chain_weight = 1

    @property
    def seed_occurrence_in_3rd_round(self) -> int:
        """:code:`bwa mem -y <int>`"""
        return self._options.max_mem_intv

    @seed_occurrence_in_3rd_round.setter
    def seed_occurrence_in_3rd_round(self, value: int) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.seed_occurrence_in_3rd_round.__name__)
        self._options.max_mem_intv = value
        self._options0.max_mem_intv = 1

    @property
    def xa_max_hits(self) -> int | tuple[int, int]:
        """:code:`bwa mem -h <int<,int>>`"""
        return self._options.max_XA_hits, self._options.max_XA_hits_alt

    @xa_max_hits.setter
    def xa_max_hits(self, value: int | tuple[int, int]) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.xa_max_hits.__name__)
        self._options0.max_XA_hits = 1
        self._options0.max_XA_hits_alt = 1
        if isinstance(value, int):
            self._options.max_XA_hits = value
            self._options.max_XA_hits_alt = value
        else:
            left, right = value
            self._options.max_XA_hits = left
            self._options.max_XA_hits_alt = right

    @property
    def xa_drop_ratio(self) -> float:
        """:code:`bwa mem -y <float>`"""
        return self._options.XA_drop_ratio

    @xa_drop_ratio.setter
    def xa_drop_ratio(self, value: float) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.xa_drop_ratio.__name__)
        self._options.XA_drop_ratio = value

    @property
    def gap_open_penalty(self) -> int | tuple[int, int]:
        """:code:`bwa mem -O <int<,int>>`"""
        if self._options.o_del == self._options.o_ins:
            return self._options.o_del
        else:
            return self._options.o_del, self._options.o_ins

    @gap_open_penalty.setter
    def gap_open_penalty(self, value: int | tuple[int, int]) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.gap_open_penalty.__name__)
        self._options0.o_del = 1
        self._options0.o_ins = 1
        if isinstance(value, int):
            self._options.o_del = value
            self._options.o_ins = value
        else:
            deletions, insertions = value
            self._options.o_del = deletions
            self._options.o_ins = insertions

    @property
    def gap_extension_penalty(self) -> int | tuple[int, int]:
        """:code:`bwa mem -E <int<,int>>`"""
        if self._options.e_del == self._options.e_ins:
            return self._options.e_del
        else:
            return self._options.e_del, self._options.e_ins

    @gap_extension_penalty.setter
    def gap_extension_penalty(self, value: int | tuple[int, int]) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.gap_extension_penalty.__name__)
        self._options0.e_del = 1
        self._options0.e_ins = 1
        if isinstance(value, int):
            self._options.e_del = value
            self._options.e_ins = value
        else:
            deletions, insertions = value
            self._options.e_del = deletions
            self._options.e_ins = insertions


    @property
    def clipping_penalty(self) -> int | tuple[int, int]:
        """:code:`bwa mem -L <int<,int>>`"""
        if self._options.pen_clip5 == self._options.pen_clip3:
            return self._options.pen_clip5
        else:
            return self._options.pen_clip5, self._options.pen_clip3

    @clipping_penalty.setter
    def clipping_penalty(self, value: int | tuple[int, int]) -> None:
        self._assert_not_finalized(attr_name=BwaMemOptions.clipping_penalty.__name__)
        self._options0.pen_clip5 = 1
        self._options0.pen_clip3 = 1
        if isinstance(value, int):
            self._options.pen_clip5 = value
            self._options.pen_clip3 = value
        else:
            five_prime, three_prime = value
            self._options.pen_clip5 = five_prime
            self._options.pen_clip3 = three_prime


cdef class BwaMem:
    """The class to align reads with :code:`bwa mem`."""

    cdef BwaIndex _index

    def __init__(self, prefix: str | Path | None = None, index: BwaIndex | None = None):
        """Constructs the :code:`bwa mem` aligner.

        One of `prefix` or `index` must be specified.

        Args:
            prefix: the path prefix for the BWA index (typically a FASTA)
            index: the index to use
        """
        if prefix is not None:
            assert Path(prefix).exists()
            self._index = BwaIndex(prefix=prefix)
        elif index is not None:
            self._index = index
        else:
            raise Exception("Either prefix or index must be given")

    # TODO: support paired end
    def align(self, queries: List[FastxRecord] | List[str], opt: BwaMemOptions | None = None) -> List[List[AlignedSegment]]:
        """Align one or more queries with `bwa aln`.

        Args:
            queries: the queries to align
            opt: the alignment options, or None to use the default options

        Returns:
            a list of alignments (:class:`~pysam.AlignedSegment`) per query
            :code:`List[List[AlignedSegment]]`.
        """
        if opt is None:
            opt = BwaMemOptions().finalize()
        elif not opt.finalized:
            opt = opt.finalize(copy=True)

        if len(queries) == 0:
            return []
        elif isinstance(queries[0], str):
            queries = [
                FastxRecord(name=f"read.{i}", sequence=sequence)
                for i, sequence in enumerate(queries)
            ]
        return self._calign(opt, queries)
    
    @staticmethod
    def __to_str(_bytes: bytes) -> str:
        return _bytes.decode('utf-8')

    cdef _copy_seq(self, q: FastxRecord, kstring_t *seq):
        seq_len = len(q.sequence)
        seq.s = <char *> calloc(sizeof(char), seq_len + 1)
        seq.l = seq_len
        seq.m = seq_len + 1
        for i, base in enumerate(q.sequence):
            seq.s[i] = nst_nt4_table[ord(base)]
        seq.s[seq_len] = b'\0'

    def _unmapped(self, query: FastxRecord) -> AlignedSegment:
        # make a default, unmapped, empty record
        rec = AlignedSegment(header=self._index.header)
        rec.query_name = query.name
        rec.reference_id = -1
        rec.reference_start = -1
        rec.mapping_quality = 0
        rec.is_paired = False
        rec.is_read1 = False
        rec.is_read2 = False
        rec.is_qcfail = False
        rec.is_duplicate = False
        rec.is_secondary = False
        rec.is_supplementary = False
        rec.is_unmapped = True
        rec.query_sequence = query.sequence
        rec.query_qualities = qualitystring_to_array(query.quality)
        return rec

    def _add_sa_tag(self, records: list[AlignedSegment]) -> None:
        num_non_secondary = sum(1 for record in records if not record.is_secondary)
        if num_non_secondary <= 1:
            return
        for i, record in enumerate(records):
            if record.is_secondary:
                continue
            # TODO: bwa mem uses the original cigar, not the one modified 
            # when not using opt.soft_clip_supplementary.  Therefore, we
            # change any hard-clip back to soft-clip
            SA = ""
            for j, other in enumerate(records):
                if i == j or other.is_secondary:
                    continue
                SA = f"{other.reference_name},{other.reference_start+1},"
                SA += "+" if other.is_forward else "-"
                SA += f",{other.cigarstring.replace('H', 'S')}"
                SA += f",{other.mapq},{other.get_tag('NM')};"
            record.set_tag("SA", SA)

    cdef _calign(self, opt: BwaMemOptions, queries: List[FastxRecord]):
        # TODO: ignore_alt
        # TODO: refactor to make this more readable
        cdef kstring_t* seqs
        cdef kstring_t* seq
        cdef char* s_char
        cdef kstring_t* kstr
        cdef mem_alnreg_v mem_alnregs
        cdef int take_all
        cdef size_t j
        cdef char **XA
        cdef mem_alnreg_t *mem_alnreg
        cdef mem_aln_t mem_aln
        cdef char *md
        cdef mem_opt_t *mem_opt

        recs_to_return: List[List[AlignedSegment]] = []

        # copy FastqProxy into bwa_seq_t
        num_seqs = len(queries)
        mem_opt = opt.mem_opt()

        seqs = <kstring_t*>calloc(sizeof(kstring_t), num_seqs)
        for i in range(num_seqs):
            seq = &seqs[i]
            query = queries[i]
            self._copy_seq(query, seq)
            mem_alnregs = mem_align1(mem_opt, self._index.bwt(), self._index.bns(), self._index.pac(), seq.l, seq.s)
            if opt.query_coord_as_primary:
                mem_reorder_primary5(opt.minimum_score, &mem_alnregs)
            # mimic mem_reg2sam from bwamem.c
            XA = NULL
            keep_all = opt.output_all_for_fragments
            if not keep_all:
                XA = mem_gen_alt(mem_opt, self._index.bns(), self._index.pac(), &mem_alnregs, seq.l, seq.s)

            mapped_recs = []
            for j in range(mem_alnregs.n):
                mem_alnreg = &mem_alnregs.a[j]
                if mem_alnreg.score < opt.minimum_score:
                    continue
                if mem_alnreg.secondary >= 0 and (mem_alnreg.is_alt > 0 or not keep_all):
                    continue
                if 0 <= mem_alnreg.secondary < INT_MAX and mem_alnreg.score < mem_alnregs.a[mem_alnreg.secondary].score * opt.xa_drop_ratio:
                    continue
                mem_aln = mem_reg2aln(mem_opt, self._index.bns(), self._index.pac(), seq.l, seq.s, mem_alnreg)
                mem_aln.XA = XA[j] if XA != NULL else NULL
                if mem_alnreg.secondary >= 0:
                    mem_aln.sub = -1  # don't output sub-optimal score
                if len(mapped_recs) > 0 and mem_alnreg.secondary < 0:  # if supplementary
                    mem_aln.flag |= 0x10000 if opt.short_split_as_secondary else 0x800
                if not opt.keep_mapq_for_supplementary and len(mapped_recs) > 0 and mem_alnreg.is_alt == 0 and mem_aln.mapq > mapped_recs[0].mapping_quality:
                    mem_aln.mapq = mapped_recs[0].mapping_quality  # lower

                rec = self._unmapped(query=query)

                # set the flags
                mem_aln.flag |= 0x4 if mem_aln.rid < 0 else 0
                mem_aln.flag |= 0x10 if mem_aln.is_rev > 0 else 0
                rec.flag = (mem_aln.flag & 0xffff) | (0x100 if (mem_aln.flag & 0x10000) != 0 else 0)
                if rec.is_unmapped:
                    continue

                # for secondary alignments, do not add sequence and qualities
                if rec.is_secondary:
                    rec.query_sequence = None
                    rec.query_qualities = None
                elif rec.is_reverse:
                    rec.query_sequence = reverse_complement(query.sequence)
                    if query.quality is not None:
                        # NB: cannot use rec.query_qualities as it is invalidated by setting
                        # query_sequence above
                        rec.query_qualities = qualitystring_to_array(query.quality[::-1])

                # reference id, position, mapq, and cigar
                rec.reference_id = mem_aln.rid
                rec.reference_start = mem_aln.pos
                rec.mapping_quality = mem_aln.mapq
                cigar = ""
                cigar_len_sum = 0
                for k in range(mem_aln.n_cigar):
                    cigar_op = mem_aln.cigar[k] & 0xf
                    if not opt.soft_clip_supplementary and mem_aln.is_alt == 0 and (
                            cigar_op == 3 or cigar_op == 4):
                        cigar_op = 4 if j > 0 else 3  # // use hard clipping for supplementary alignments
                    cigar_len = mem_aln.cigar[k] >> 4
                    cigar += f"{cigar_len}" + "MIDSH"[cigar_op]
                    if cigar_op < 4:
                        cigar_len_sum += cigar_len
                rec.cigarstring = cigar

                # remove leading and trailing soft-clipped bases for non-primary etc.
                if mem_aln.n_cigar > 0 and j > 0 and not opt.soft_clip_supplementary and mem_aln.is_alt == 0:
                    qb = 0
                    qe = len(query.sequence)
                    leading_op = mem_aln.cigar[0] & 0xf
                    trailing_op = mem_aln.cigar[mem_aln.n_cigar - 1] & 0xf
                    if leading_op == 3 or leading_op == 4:
                        qb += mem_aln.cigar[0] >> 4
                    if trailing_op == 3 or trailing_op == 4:
                        qe -= mem_aln.cigar[mem_aln.n_cigar - 1] >> 4
                    query_qualities = rec.query_qualities # as setting query_sequence may invalidate this
                    if rec.query_sequence is not None:
                        rec.query_sequence = rec.query_sequence[qb:qe]
                    if query_qualities is not None:
                        rec.query_qualities = query_qualities[qb:qe]

                # Optional tags
                attrs = dict()
                if mem_aln.n_cigar > 0:
                    attrs["NM"] = mem_aln.NM
                    md = <char *> (mem_aln.cigar + mem_aln.n_cigar)
                    attrs["MD"] = self.__to_str(md)
                # NB: mate tags are not output: MC, MQ
                if mem_aln.score >= 0:
                    attrs["AS"] = mem_aln.score
                if mem_aln.sub >= 0:
                    attrs["XS"] = mem_aln.sub
                # NB: SA is added after all the records have been created
                if mem_aln.XA != NULL:
                    attrs["XB" if opt.with_xb_tag else "XA"] = mem_aln.XA
                if opt.with_xr_tag and self._index.bns().anns[rec.reference_id].anno != 0 and \
                        self._index.bns().anns[rec.reference_id].anno[0] != 0:
                    attrs["XR"] = str(self._index.bns().anns[rec.reference_id].anno)
                rec.set_tags(list(attrs.items()))

                mapped_recs.append(rec)

                free(mem_aln.cigar)
            if len(mapped_recs) == 0:
                recs_to_return.append([self._unmapped(query=query)])
            else:
                self._add_sa_tag(mapped_recs)
                recs_to_return.append(mapped_recs)

            if XA != NULL:
                for j in range(len(mapped_recs)):
                    free(XA[j])
                free(XA)
            free(mem_alnregs.a)

        for i in range(num_seqs):
            free(seqs[i].s)
        free(seqs)

        return recs_to_return



# cython: language_level=3
from pathlib import Path

from libc.stdint cimport int64_t, int32_t, uint8_t
from libc.stdio cimport FILE

cdef extern from "bwa.h":
    char * bwa_idx_infer_prefix(const char * hint)
    ctypedef struct bwaidx_t:
        bwt_t    *bwt # FM-index
        bntseq_t *bns # information on the reference sequences
        uint8_t  *pac # the actual 2-bit encoded reference sequences with 'N' converted to a random base

    bwaidx_t * bwa_idx_load(const char * hint, int which)
    void bwa_idx_destroy(bwaidx_t *idx)
    int BWA_IDX_BWT
    int BWA_IDX_BNS
    int BWA_IDX_PAC
    int bwa_idx_build(const char *fa, const char *prefix, int algo_type, int block_size)


cdef extern from "bwt.h":
    ctypedef struct bwt_t:
        int sa_intv

    bwt_t *bwt_restore_bwt(const char *fn)
    void bwt_restore_sa(const char *fn, bwt_t *bwt);
    void bwt_destroy(bwt_t *bwt)

cdef extern from "bntseq.h":
    ctypedef  struct bntann1_t:
        int64_t offset
        int32_t len
        char *name

    ctypedef struct bntseq_t:
        int64_t l_pac
        bntann1_t *anns
        FILE * fp_pac

    bntseq_t * bns_restore(const char * prefix)
    void bns_destroy(bntseq_t *bns)

cdef bytes force_bytes(object s)

cdef class BwaIndex:
    """Contains the index and nucleotide sequence for Bwa"""
    cdef bwaidx_t *_delegate
    cdef public object header
    cdef bwt_t *bwt(self)
    cdef bntseq_t *bns(self)
    cdef uint8_t *pac(self)
    cdef _load_index(self, prefix, mode)



# cython: language_level=3

from pathlib import Path

from cpython cimport PyBytes_Check, PyUnicode_Check
from pysam import AlignmentFile, FastxFile
import enum


__all__ = [
    "BwaIndex",
    "BwaIndexBuildMethod",
]

@enum.unique
class BwaIndexBuildMethod(enum.Enum):
    """The BWT construction algorithm (:code:`bwa index -a <str>`)"""

    AUTO = enum.auto()
    RB2 = enum.auto()
    BWTSW = enum.auto()
    IS = enum.auto()


cdef str ERROR_HANDLER = 'strict'
cdef str TEXT_ENCODING = 'utf-8'


cdef bytes force_bytes(object s):
    return force_bytes_with(s, None, None)


cdef bytes force_bytes_with(object s, encoding: str | None = None, errors: str | None = None):
    """convert string or unicode object to bytes, assuming
    utf8 encoding.
    """
    if s is None:
        return None
    elif PyBytes_Check(s):
        return s
    elif PyUnicode_Check(s):
        return s.encode(encoding or TEXT_ENCODING, errors or ERROR_HANDLER)
    else:
        raise TypeError("Argument must be string, bytes or unicode.")



cdef class BwaIndex:
    """Contains the index and nucleotide sequence for Bwa.  Use :code:`bwa index` on the command
    line to generate the bwa index.

    Note: the accompanying sequence dictionary must exist (i.e. `.dict` file, which can be generated
    with :code:`samtools dict <fasta>`).

    Args:
        prefix (str | Path): the path prefix for the BWA index (typically a FASTA)
        bwt (bool): load the BWT (FM-index)
        bns (bool): load the BNS (reference sequence metadata)
        pac (bool): load the PAC (the actual 2-bit encoded reference sequences with 'N' converted to a
             random base)
    """

    def __init__(self, prefix: str | Path, bwt: bool = True, bns: bool = True, pac: bool = True) -> None:
        cdef int mode

        mode = 0
        if bwt:
            mode |= BWA_IDX_BWT
        if bns:
            mode |= BWA_IDX_BNS
        if pac:
            mode |= BWA_IDX_PAC
        self._load_index(f"{prefix}", mode)

    @classmethod
    def index(cls,
              fasta: str | Path,
              method: BwaIndexBuildMethod= BwaIndexBuildMethod.AUTO,
              prefix: str | Path | None = None,
              block_size: int = 10000000,
              out_64: bool = False) -> None:
        """Indexes a given FASTA.  Also builds the sequence dictionary (.dict).

        Args:
            fasta (str | Path): the path to the FASTA to index
            method (BwaIndexBuildMethod): the BWT construction algorithm (:code:`bwa index -a <str>`)
            prefix (str | Path | None): the path prefix for the BWA index (typically a FASTA)
            block_size (int): block size for the bwtsw algorithm (effective with -a bwtsw)
            out_64 (bool): index files named as :code:`<in.fasta>.64.*` instead of :code:`<in.fasta>.*`
        """
        if prefix is None:
            prefix = fasta

        # Build the BWA index
        bwa_idx_build(force_bytes(f"{fasta}"), force_bytes(f"{prefix}"), method.value, block_size)

        # Build the sequence dictionary
        dict_fn = Path(prefix.with_suffix(".dict"))
        with FastxFile(filename=f"{fasta}") as reader, dict_fn.open("w") as writer:
            writer.write("@HD\tVN:1.5\tSO:unsorted\n")
            for rec in reader:
                writer.write(f"@SQ\tSN:{rec.name}\tLN:{len(rec.sequence)}\n")


    cdef _load_index(self, prefix, mode):
        prefix = bwa_idx_infer_prefix(force_bytes(prefix))
        if not prefix:
            raise Exception(f"Could not find the index at: {prefix}")

        self._delegate = bwa_idx_load(prefix, mode)

        # the SAM header from the sequence dictionary
        seq_dict = Path(prefix.decode("utf-8")).with_suffix(".dict")
        # TODO: error message when seq_dict is missing?
        with seq_dict.open("r") as fh:
            with AlignmentFile(fh) as reader:
                self.header = reader.header

    cdef bwt_t *bwt(self):
        return self._delegate.bwt

    cdef bntseq_t *bns(self):
        return self._delegate.bns

    cdef uint8_t *pac(self):
        return self._delegate.pac

    def __dealloc__(self):
        bwa_idx_destroy(self._delegate)
        self._delegate = NULL



from pathlib import Path
from typing import List

from pysam import AlignedSegment
from pysam import FastxRecord

from pybwa.libbwaindex import BwaIndex

class BwaAlnOptions:
    def __init__(
        self,
        max_mismatches: int | None = None,
        max_gap_opens: int | None = None,
        max_gap_extensions: int | None = None,
        min_indel_to_end_distance: int | None = None,
        max_occurrences_for_extending_long_deletion: int | None = None,
        seed_length: int | None = None,
        max_mismatches_in_seed: int | None = None,
        mismatch_penalty: int | None = None,
        gap_open_penalty: int | None = None,
        gap_extension_penalty: int | None = None,
        stop_at_max_best_hits: int | None = None,
        max_hits: int | None = 3,
        log_scaled_gap_penalty: bool | None = None,
    ) -> None: ...
    max_mismatches: int  # -n <int>
    # fnr:float # -n <float>
    max_gap_opens: int  # -o <int>
    max_gap_extensions: int  # -e <int>
    min_indel_to_end_distance: int  # -i <int>
    max_occurrences_for_extending_long_deletion: int  # -d <int>
    seed_length: int  # -l <int>
    max_mismatches_in_seed: int  # -k <int>
    mismatch_penalty: int  # -M <int>
    gap_open_penalty: int  # -O <int>
    gap_extension_penalty: int  # -E <int>
    stop_at_max_best_hits: int  # -R <int>
    max_hits: int  # bwa samse -n <int>
    log_scaled_gap_penalty: bool = True  # -L

class BwaAln:
    def __init__(self, prefix: str | Path | None = None, index: BwaIndex | None = None) -> None: ...
    def align(
        self, queries: List[FastxRecord] | List[str], opt: BwaAlnOptions | None = None
    ) -> List[AlignedSegment]: ...
